#
#     Function file: functions.s
#
#     Example 4.1: Randomized block design
#     Example 4.2: Heteroscedastic errors
#
#     =======================================================
#
#     Compute B-spline base matrix
#
#     x = x-domain vector
#     xl = left of x-domain, xr = right of x-domain
#     ndx = number of intervals in domain
#     bdeg = degree of B-spline (quadratic = 2, etc)
#
bspline <- function(x, xl, xr, ndx, bdeg){
 dx <- (xr-xl)/ndx
 knots <- seq(xl-bdeg*dx, xr+bdeg*dx, by=dx)
 B <- spline.des(knots,x,bdeg+1,0*x)$design
 B
}
#
#       Function to construct a block diagonal matrix from
#       a list of matrices or vectors.  x is a list of matrices.
#       NULLs and scalars are allowed in the list.
#   
dbind <- function(x){
  len.x <- length(x)
  dim.x <- do.call("cbind", lapply(x, function(x)
  if(length(x) == 0) 0 else {
    if(length(x) == 1) c(1, 1) else dim(x)
  }
  ))	
  dim.res <- c( sum(dim.x[1,  ]), sum(dim.x[2,  ]) )
  res <- array(0, dim.res)
  res.ends <- cumsum(dim.x[1,  ])
  res.ends2 <- cumsum(dim.x[2,  ])
  res.starts <- c(1, res.ends[ - len.x] + 1)
  res.starts2 <- c(1, res.ends2[ - len.x] + 1)
  for(i in seq(len.x))
  if(dim.x[1, i] > 0) res[res.starts[i]:res.ends[i],
	                  res.starts2[i]:res.ends2[i]] <- x[[i]]
  res
}
#
#     Compute augmented matrix
#
Augment <- function(M1., M2.){ 
  n <- ncol(M1.)
  Zero <- matrix(0, nrow = n, ncol = n)
  dbind( list(Zero, M2.) )
}
#
#     Function to calculate AIC.c, \hat a, yhat, S2 and tr
#
#     y = response
#     lambda = penalty weight
#     Mat.1 is matrix [X:B]
#     !
#     !   Care! The columns of X corresponding to the block effects
#     !   must be deleted.  
#     !
#     B is block diagonal Bspline basis
#     Mat.2 is augmented D'D matrix where D'D is block diagonal
#
Calc <- function(Mat.1, Mat.2, y, lambda){
  MtM <- t(Mat.1) %*% Mat.1
  Q <- solve(MtM + lambda * Mat.2)
  a <- Q %*% t(Mat.1) %*% y
  yhat <- Mat.1 %*% a 
  n <- length(y)
  s2 <- sum((y - yhat)^2)/n
  tr <- sum( diag(Q %*% MtM) )
  aic.c <- log(s2) + 2 * (tr+1)/(n-tr-2)
  A <- list(Aic.c = aic.c, Coef = a, Fitted = yhat, Sigma2 = s2, 
     Trace = tr) 
  A
}
#
#     AIC.c for distinct $\lambda$
#     Mat.1 is XpB, Mat.2 is DtD, lambda is (l1, l2, l3)
#
Calc.3 <- function(Mat.1, Mat.2, y, X, lambda){
  MtM <- t(Mat.1) %*% Mat.1
  Mat <- dbind(lambda[1] * Mat.2, lambda[2] * Mat.2, lambda[3] * Mat.2)
  Mat <- Augment(X, Mat)
  Q <- solve(MtM + Mat)
  a <- Q %*% t(Mat.1) %*% y
  yhat <- Mat.1 %*% a 
  n <- length(y)
  s2 <- sum((y - yhat)^2)/n
  tr <- sum( diag(Q %*% MtM) )
  aic.c <- log(s2) + 2 * (tr+3)/(n-tr-4)
  A <- list(Aic.c = aic.c, Coef = a, Fitted = yhat, Sigma2 = s2, 
     Trace = tr) 
  A
}
#
#     =============================================================
#
#     REML functions
#
#     Define determinant function
#
det <- function(M) prod(svd(M, nu=0, nv=0)$d)
#
#    QR function to compute R00, R11.1 and cm1 in 
#    Pinheiro and Bates p 76, equation (2.23) 
#
QR <- function(X., Z., Y., Lambda.){
  q <- ncol(Z.)
  ytil <- c(Y., rep(0, q))
  Xtil <- rbind(X., matrix(0, nrow = q, ncol = ncol(X.)))
  Ztil <- rbind(Z., sqrt(Lambda.) * diag(q))
  Ztil.qr <- qr( Ztil )
  R11.1 <- qr.R(Ztil.qr)[1:q, 1:q]
  QtXy <- qr.qty(Ztil.qr, cbind(Xtil, ytil) )
  R00.1 <- QtXy[(q+1):nrow(QtXy), 1:(ncol(QtXy) - 1) ]
  c0.1 <- QtXy[(q+1):nrow(QtXy), ncol(QtXy) ]
  Rc.qr <- qr( cbind(R00.1, c0.1) )
  R <- qr.R( Rc.qr )
  k <- ncol(R) - 1
  R00 <- R[1:k, 1:k]
  cm1 <- R[(k+1):nrow(R), (k+1)]
  return(R11.1, R00, cm1)
}
#
#     Evaluate the profile REML via QR
#     Pinheiro and Bates, p76
#
P.Reml.QR <- function(X., Z., Y., Lambda.){
  qr <- QR(X., Z., Y., Lambda.)
  n <- length(Y.)
  p <- ncol(X.)
  l1 <- -0.5*(n-p)*log(sum(qr$cm1^2))
  l2 <- -log(abs(det(qr$R00)))
  l3 <- log( det(sqrt(Lambda.)*diag(ncol(Z.)))/det(abs(qr$R11.1)) )
  l1 + l2 + l3
}
#
#     Evaluate conditional estimate of sigma2
#
CondS2 <- function(X., Z., Y., Lambda.){
  QR(X., Z., Y., Lambda.)$cm1^2/(length(Y.) - ncol(X.))
}
#
#     And for separate lambda
#
P.Reml.3 <- function(X., Z., Y., L.){
  bigZ <- dbind(list( Z./sqrt(L.[1]), Z./sqrt(L.[2]), Z./sqrt(L.[3]) ))
  P.Reml.QR(X., bigZ, Y., 1)
}
#
#     ======================================================
#
#     Plotting functions: residuals and fitted smooths
#
#     plot.1 draws a single panel in LH plot of Fig 2
#
#     y  = yield in a block
#     r1 = residuals from variety and wheel (common lambda)
#     s1 = smooth (common lambda)
#     r3 = residuals from variety and wheel (separate lambda)
#     s3 = smooth (separate lambda)
#     b1 = blocksize
#     n  = block index
#
plot.1 <- function(y, r1, s1, r3, s3, b1, n){
  centre <- mean(y - r1)
  y.plot <- y - r1 - centre
  Xplot <- c(0.5, rep(1.5:(b1-0.5), rep(2,(b1-1))), b1+0.5)
  Yplot <- rep(y.plot, rep(2,b1))
  plot(Xplot, Yplot, type = "l", xlab = "", ylab = paste("Block", n), 
  xaxt = "n", cex = 0.8)
  lines(1:b1, s1 - centre)
  centre.3 <- mean( y - r3)
  lines(1:b1, s3 - centre.3, lty = 2)
}
#
#      plot 3 calls plot.1 for each block
#
plot3 <- function(y, r1, s1, r3, s3, b1){
  par(mfcol = c(3,2))
  par(mar = c(5,6,1,1))
  R1 <- 1:b1
  plot.1(y[R1], r1[R1], s1[R1], r3[R1], s3[R1], b1, 1)
  R2 <- (b1+1):(2*b1)
  plot.1(y[R2], r1[R2], s1[R2], r3[R2], s3[R2], b1, 2)
  R3 <- (2*b1+1):(3*b1)
  plot.1(y[R3], r1[R3], s1[R3], r3[R3], s3[R3], b1, 3)
}
#
#     Plotting functions: standardized residuals
#
#     plot.1.res draws a single panel in RH plot of Fig 2
#     Parameters as above - sd = \hat\sigma
#
plot.1.res <- function(y, r1, s1, sd, b1, n){
  y.plot <- (y - r1 - s1)/sd
  Xplot <- c(0.5, rep(1.5:(b1-0.5), rep(2,(b1-1))), b1+0.5)
  Yplot.1 <- rep(y.plot, rep(2,b1))
  plot(Xplot, Yplot.1, type = "l", xlab = "", ylab = paste("Block", n), 
  xaxt = "n", cex = 0.8)
}
#
#      plot3.res calls plot.1.res for each block
#
plot3.res <- function(y, r1, s1, sd, b1){
    par(mar = c(5,6,1,1))
    R1 <- 1:b1
    plot.1.res(y[R1], r1[R1], s1[R1], sd, b1, 1)
    R2 <- (b1+1):(2*b1)
    plot.1.res(y[R2], r1[R2], s1[R2], sd, b1, 2)
    R3 <- (2*b1+1):(3*b1)
    plot.1.res(y[R3], r1[R3], s1[R3], sd, b1, 3)
}


